"""
Palma construction catalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Cecilia Pérez Pérez cperez@irec.cat
"""

import json
from pathlib import Path
from hub.catalog_factories.catalog import Catalog
from hub.catalog_factories.data_models.construction.content import Content
from hub.catalog_factories.construction.construction_helper import ConstructionHelper
from hub.catalog_factories.data_models.construction.construction import Construction
from hub.catalog_factories.data_models.construction.archetype import Archetype
from hub.catalog_factories.data_models.construction.window import Window
from hub.catalog_factories.data_models.construction.material import Material
from hub.catalog_factories.data_models.construction.layer import Layer
import hub.helpers.constants as cte


class PalmaCatalog(Catalog):
  """
  Palma catalog class
  """
  def __init__(self, path):
    _path_archetypes = Path(path / 'palma_archetypes.json').resolve()
    _path_constructions = (path / 'palma_constructions.json').resolve()
    with open(_path_archetypes, 'r', encoding='utf-8') as file:
      self._archetypes = json.load(file)
    with open(_path_constructions, 'r', encoding='utf-8') as file:
      self._constructions = json.load(file)

    self._catalog_windows = self._load_windows()
    self._catalog_materials = self._load_materials()
    self._catalog_constructions = self._load_constructions()
    self._catalog_archetypes = self._load_archetypes()

    # store the full catalog data model in self._content
    self._content = Content(self._catalog_archetypes,
                            self._catalog_constructions,
                            self._catalog_materials,
                            self._catalog_windows)

  def _load_windows(self):
    _catalog_windows = []
    windows = self._constructions['transparent_surfaces']
    for window in windows:
      name = list(window.keys())[0]
      window_id = name
      g_value = window[name]['shgc']
      window_type = window[name]['type']
      frame_ratio = window[name]['frame_ratio']
      overall_u_value = window[name]['u_value']
      _catalog_windows.append(Window(window_id, frame_ratio, g_value, overall_u_value, name, window_type))
    return _catalog_windows

  def _load_materials(self):
    _catalog_materials = []
    materials = self._constructions['materials']
    for material in materials:
      name = list(material.keys())[0]
      material_id = name
      no_mass = material[name]['no_mass']
      thermal_resistance = None
      conductivity = None
      density = None
      specific_heat = None
      solar_absorptance = None
      thermal_absorptance = None
      visible_absorptance = None
      if no_mass:
        thermal_resistance = material[name]['thermal_resistance']
      else:
        solar_absorptance = material[name]['solar_absorptance']
        thermal_absorptance = str(1 - float(material[name]['thermal_emittance']))
        visible_absorptance = material[name]['visible_absorptance']
        conductivity = material[name]['conductivity']
        density = material[name]['density']
        specific_heat = material[name]['specific_heat']
      _material = Material(material_id,
                           name,
                           solar_absorptance,
                           thermal_absorptance,
                           visible_absorptance,
                           no_mass,
                           thermal_resistance,
                           conductivity,
                           density,
                           specific_heat)
      _catalog_materials.append(_material)
    return _catalog_materials

  def _load_constructions(self):
    _catalog_constructions = []
    constructions = self._constructions['opaque_surfaces']
    for construction in constructions:
      name = list(construction.keys())[0]
      construction_id = name
      construction_type = ConstructionHelper().nrcan_surfaces_types_to_hub_types[construction[name]['type']]
      layers = []
      for layer in construction[name]['layers']:
        layer_id = layer
        layer_name = layer
        material_id = layer
        thickness = construction[name]['layers'][layer]
        for material in self._catalog_materials:
          if str(material_id) == str(material.id):
            layers.append(Layer(layer_id, layer_name, material, thickness))
            break
      _catalog_constructions.append(Construction(construction_id, construction_type, name, layers))
    return _catalog_constructions

  def _load_archetypes(self):
    _catalog_archetypes = []
    archetypes = self._archetypes['archetypes']
    for archetype in archetypes:
      archetype_id = f'{archetype["function"]}_{archetype["period_of_construction"]}_{archetype["climate_zone"]}'
      function = archetype['function']
      name = archetype_id
      climate_zone = archetype['climate_zone']
      construction_period = archetype['period_of_construction']
      average_storey_height = archetype['average_storey_height']
      thermal_capacity = float(archetype['thermal_capacity']) * 1000
      extra_loses_due_to_thermal_bridges = archetype['extra_loses_due_thermal_bridges']
      infiltration_rate_for_ventilation_system_off = archetype['infiltration_rate_for_ventilation_system_off'] / cte.HOUR_TO_SECONDS
      infiltration_rate_for_ventilation_system_on = archetype['infiltration_rate_for_ventilation_system_on'] / cte.HOUR_TO_SECONDS
      infiltration_rate_area_for_ventilation_system_off = (
              archetype['infiltration_rate_area_for_ventilation_system_off'] * 1
      )
      infiltration_rate_area_for_ventilation_system_on = (
              archetype['infiltration_rate_area_for_ventilation_system_on'] * 1
      )

      archetype_constructions = []
      for archetype_construction in archetype['constructions']:
        archetype_construction_type = ConstructionHelper().nrcan_surfaces_types_to_hub_types[archetype_construction]
        archetype_construction_name = archetype['constructions'][archetype_construction]['opaque_surface_name']
        for construction in self._catalog_constructions:
          if archetype_construction_type == construction.type and construction.name == archetype_construction_name:
            _construction = None
            _window = None
            _window_ratio = None
            if 'transparent_surface_name' in archetype['constructions'][archetype_construction].keys():
              _window_ratio = archetype['constructions'][archetype_construction]['transparent_ratio']
              _window_id = archetype['constructions'][archetype_construction]['transparent_surface_name']
              for window in self._catalog_windows:
                if _window_id == window.id:
                  _window = window
                  break
            _construction = Construction(construction.id,
                                         construction.type,
                                         construction.name,
                                         construction.layers,
                                         _window_ratio,
                                         _window)
            archetype_constructions.append(_construction)
            break

      _catalog_archetypes.append(Archetype(archetype_id,
                                           name,
                                           function,
                                           climate_zone,
                                           construction_period,
                                           archetype_constructions,
                                           average_storey_height,
                                           thermal_capacity,
                                           extra_loses_due_to_thermal_bridges,
                                           None,
                                           infiltration_rate_for_ventilation_system_off,
                                           infiltration_rate_for_ventilation_system_on,
                                           infiltration_rate_area_for_ventilation_system_off,
                                           infiltration_rate_area_for_ventilation_system_on))
    return _catalog_archetypes

  def names(self, category=None):
    """
    Get the catalog elements names
    :parm: optional category filter
    """
    if category is None:
      _names = {'archetypes': [], 'constructions': [], 'materials': [], 'windows': []}
      for archetype in self._content.archetypes:
        _names['archetypes'].append(archetype.name)
      for construction in self._content.constructions:
        _names['constructions'].append(construction.name)
      for material in self._content.materials:
        _names['materials'].append(material.name)
      for window in self._content.windows:
        _names['windows'].append(window.name)
    else:
      _names = {category: []}
      if category.lower() == 'archetypes':
        for archetype in self._content.archetypes:
          _names[category].append(archetype.name)
      elif category.lower() == 'constructions':
        for construction in self._content.constructions:
          _names[category].append(construction.name)
      elif category.lower() == 'materials':
        for material in self._content.materials:
          _names[category].append(material.name)
      elif category.lower() == 'windows':
        for window in self._content.windows:
          _names[category].append(window.name)
      else:
        raise ValueError(f'Unknown category [{category}]')
    return _names

  def entries(self, category=None):
    """
    Get the catalog elements
    :parm: optional category filter
    """
    if category is None:
      return self._content
    if category.lower() == 'archetypes':
      return self._content.archetypes
    if category.lower() == 'constructions':
      return self._content.constructions
    if category.lower() == 'materials':
      return self._content.materials
    if category.lower() == 'windows':
      return self._content.windows
    raise ValueError(f'Unknown category [{category}]')

  def get_entry(self, name):
    """
    Get one catalog element by names
    :parm: entry name
    """
    for entry in self._content.archetypes:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.constructions:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.materials:
      if entry.name.lower() == name.lower():
        return entry
    for entry in self._content.windows:
      if entry.name.lower() == name.lower():
        return entry
    raise IndexError(f"{name} doesn't exists in the catalog")


"""
TestMontrealCustomCatalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Atiya atiya.atiya@mail.concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from unittest import TestCase
from hub.catalog_factories.costs_catalog_factory import CostsCatalogFactory


class TestCostsCatalog(TestCase):

  def test_costs_catalog(self):
    catalog = CostsCatalogFactory('montreal_custom').catalog
    catalog_categories = catalog.names()
    self.assertIsNotNone(catalog, 'catalog is none')
    content = catalog.entries()
    self.assertTrue(len(content.archetypes) == 2)

    # retrieving all the entries should not raise any exceptions
    for category in catalog_categories:
      for value in catalog_categories[category]:
        catalog.get_entry(value)

    with self.assertRaises(IndexError):
      catalog.get_entry('unknown')



"""
export a city into Geojson format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""
import json
from pathlib import Path

import numpy as np
import pyproj
from pyproj import Transformer

from hub.helpers.geometry_helper import GeometryHelper


class Geojson:
  """
  Export to geojson format
  """
  def __init__(self, city, path, target_buildings):
    self._city = city
    self._file_path = Path(path / f'{self._city.name}.geojson').resolve()
    try:
      srs_name = self._city.srs_name
      if self._city.srs_name in GeometryHelper.srs_transformations:
        srs_name = GeometryHelper.srs_transformations[self._city.srs_name]
      input_reference = pyproj.CRS(srs_name)  # Projected coordinate system from input data
    except pyproj.exceptions.CRSError as err:
      raise pyproj.exceptions.CRSError from err
    self._to_gps = Transformer.from_crs(input_reference, pyproj.CRS('EPSG:4326'))
    if target_buildings is None:
      target_buildings = [b.name for b in self._city.buildings]
    self._geojson_skeleton = {
      'type': 'FeatureCollection',
      'features': []
    }
    self._feature_skeleton = {
      'type': 'Feature',
      'geometry': {
        'type': 'Polygon',
        'coordinates': []
      },
      'properties': {}
    }
    self._export()

  def _export(self):
    for building in self._city.buildings:
      if len(building.grounds) == 1:
        ground = building.grounds[0]
        feature = self._polygon(ground)
      else:
        feature = self._multipolygon(building.grounds)
      feature['id'] = building.name
      feature['properties']['height'] = f'{building.max_height - building.lower_corner[2]}'
      feature['properties']['function'] = f'{building.function}'
      feature['properties']['year_of_construction'] = f'{building.year_of_construction}'
      feature['properties']['aliases'] = building.aliases
      feature['properties']['elevation'] = f'{building.lower_corner[2]}'
      self._geojson_skeleton['features'].append(feature)
    with open(self._file_path, 'w', encoding='utf-8') as f:
      json.dump(self._geojson_skeleton, f, indent=2)

  def _polygon(self, ground):
    feature = {
      'type': 'Feature',
      'geometry': {
        'type': 'Polygon',
        'coordinates': []
      },
      'properties': {}
    }
    ground_coordinates = []
    for coordinate in ground.solid_polygon.coordinates:
      gps_coordinate = self._to_gps.transform(coordinate[0], coordinate[1])
      ground_coordinates.insert(0, [gps_coordinate[1], gps_coordinate[0]])

    first_gps_coordinate = self._to_gps.transform(
      ground.solid_polygon.coordinates[0][0],
      ground.solid_polygon.coordinates[0][1]
    )
    ground_coordinates.insert(0, [first_gps_coordinate[1], first_gps_coordinate[0]])
    feature['geometry']['coordinates'].append(ground_coordinates)
    return feature

  def _multipolygon(self, grounds):
    feature = {
      'type': 'Feature',
      'geometry': {
        'type': 'MultiPolygon',
        'coordinates': []
      },
      'properties': {}
    }
    polygons = []
    for ground in grounds:
      ground_coordinates = []
      for coordinate in ground.solid_polygon.coordinates:
        gps_coordinate = self._to_gps.transform(coordinate[0], coordinate[1])
        ground_coordinates.insert(0, [gps_coordinate[1], gps_coordinate[0]])

      first_gps_coordinate = self._to_gps.transform(
        ground.solid_polygon.coordinates[0][0],
        ground.solid_polygon.coordinates[0][1]
      )
      ground_coordinates.insert(0, [first_gps_coordinate[1], first_gps_coordinate[0]])
      polygons.append(ground_coordinates)
    feature['geometry']['coordinates'].append(polygons)
    return feature





"""
Energy Systems catalog factory, publish the energy systems information
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Álvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from pathlib import Path
from typing import TypeVar

from hub.catalog_factories.energy_systems.montreal_custom_catalog import MontrealCustomCatalog
from hub.catalog_factories.energy_systems.montreal_future_system_catalogue import MontrealFutureSystemCatalogue
from hub.catalog_factories.energy_systems.palma_system_catalgue import PalmaSystemCatalogue
from hub.helpers.utils import validate_import_export_type

Catalog = TypeVar('Catalog')


class EnergySystemsCatalogFactory:
  """
  Energy system catalog factory class
  """
  def __init__(self, handler, base_path=None):
    if base_path is None:
      base_path = Path(Path(__file__).parent.parent / 'data/energy_systems')
    self._handler = '_' + handler.lower()
    validate_import_export_type(EnergySystemsCatalogFactory, handler)
    self._path = base_path

  @property
  def _montreal_custom(self):
    """
    Retrieve NRCAN catalog
    """
    return MontrealCustomCatalog(self._path)

  @property
  def _montreal_future(self):
    """
    Retrieve North American catalog
    """
    return MontrealFutureSystemCatalogue(self._path)

  @property
  def _palma(self):
    """
    Retrieve Palma catalog
    """
    return PalmaSystemCatalogue(self._path)

  @property
  def catalog(self) -> Catalog:
    """
    Enrich the city given to the class using the class given handler
    :return: Catalog
    """
    return getattr(self, self._handler, lambda: None)



"""
weather helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
import math
import hub.helpers.constants as cte
from datetime import datetime, timedelta

class Weather:
  """
  Weather class
  """

  _epw_file = {
    'CA.02.5935': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/BC/CAN_BC_Summerland.717680_CWEC/CAN_BC_Summerland.717680_CWEC.epw',
    'CA.10.06': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'CA.10.13': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'CA.10.14': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'CA.10.16': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Montreal.Intl.AP.716270_CWEC/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw',
    'DE.01.082': 'https://energyplus-weather.s3.amazonaws.com/europe_wmo_region_6/DEU/DEU_Stuttgart.107380_IWEC/DEU_Stuttgart.107380_IWEC.epw',
    'US.NY.047': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/USA/NY/USA_NY_New.York.City-Central.Park.94728_TMY/USA_NY_New.York.City-Central.Park.94728_TMY.epw',
    'CA.10.12': 'https://energyplus-weather.s3.amazonaws.com/north_and_central_america_wmo_region_4/CAN/PQ/CAN_PQ_Quebec.717140_CWEC/CAN_PQ_Quebec.717140_CWEC.epw',
    'IL.01.': 'https://energyplus-weather.s3.amazonaws.com/europe_wmo_region_6/ISR/ISR_Eilat.401990_MSI/ISR_Eilat.401990_MSI.epw',
    'ES.07.PM': 'https://energyplus-weather.s3.amazonaws.com/europe_wmo_region_6/ESP/ESP_Palma.083060_SWEC/ESP_Palma.083060_SWEC.epw'
  }
  # todo: this dictionary need to be completed, a data science student task?

  @staticmethod
  def sky_temperature(ambient_temperature):
    """
    Get sky temperature from ambient temperature in Celsius
    :return: List[float]
    """
    # Swinbank - Source sky model approximation(1963) based on cloudiness statistics(32 %) in the United States
    # ambient temperatures( in °C)
    # sky temperatures( in °C)
    values = []
    for temperature in ambient_temperature:
      value = 0.037536 * math.pow((temperature + cte.KELVIN), 1.5) \
              + 0.32 * (temperature + cte.KELVIN) - cte.KELVIN
      values.append(value)
    return values

  @staticmethod
  def cold_water_temperature(ambient_temperature):
    """
    Get cold water temperature from ambient temperature in Celsius
    :return: dict
    """
    # Equation from "TOWARDS DEVELOPMENT OF AN ALGORITHM FOR MAINS WATER TEMPERATURE", 2004, Jay Burch
    # and Craig Christensen, National Renewable Energy Laboratory
    # ambient temperatures( in °C)
    # cold water temperatures( in °C)
    t_out_fahrenheit = [1.8 * t_out + 32 for t_out in ambient_temperature]
    t_out_average = sum(t_out_fahrenheit) / len(t_out_fahrenheit)
    max_difference = max(t_out_fahrenheit) - min(t_out_fahrenheit)
    ratio = 0.4 + 0.01 * (t_out_average - 44)
    lag = 35 - (t_out_average - 35)
    number_of_day = [a for a in range(1, 366)]
    day_of_year = [day for day in number_of_day for _ in range(24)]
    cold_temperature_fahrenheit = []
    cold_temperature = []
    for i in range(len(ambient_temperature)):
      cold_temperature_fahrenheit.append(t_out_average + 6 + ratio * (max_difference / 2) *
                                         math.sin(math.radians(0.986 * (day_of_year[i] - 15 - lag) - 90)))
      cold_temperature.append((cold_temperature_fahrenheit[i] - 32) / 1.8)
    return cold_temperature

  def epw_file(self, region_code):
    """
    returns the url for the weather file for the given location or default (Montreal data)
    :return: str
    """
    if region_code not in self._epw_file:
      logging.warning('Specific weather data unknown for %s using Montreal data instead', region_code)
      return self._epw_file['CA.10.06']
    return self._epw_file[region_code]



"""
Model representation of an application
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""

import datetime

from sqlalchemy import Column, Integer, String, Sequence
from sqlalchemy import DateTime
from sqlalchemy.dialects.postgresql import UUID

from hub.persistence.configuration import Models


class Application(Models):
  """
  A model representation of an application
  """
  __tablename__ = 'application'
  id = Column(Integer, Sequence('application_id_seq'), primary_key=True)
  name = Column(String, nullable=False)
  description = Column(String, nullable=False)
  application_uuid = Column(UUID(as_uuid=True), nullable=False)
  created = Column(DateTime, default=datetime.datetime.utcnow)
  updated = Column(DateTime, default=datetime.datetime.utcnow)

  def __init__(self, name, description, application_uuid):
    self.name = name
    self.description = description
    self.application_uuid = application_uuid



"""
ThermalBoundary module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import uuid
import math
from typing import List, Union, TypeVar
import logging
from hub.helpers.configuration_helper import ConfigurationHelper as ch
import hub.helpers.constants as cte
from hub.city_model_structure.building_demand.layer import Layer
from hub.city_model_structure.building_demand.thermal_opening import ThermalOpening
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone

Surface = TypeVar('Surface')


class ThermalBoundary:
  """
  ThermalBoundary class
  """
  def __init__(self, parent_surface, opaque_area, windows_areas):
    self._parent_surface = parent_surface
    self._opaque_area = opaque_area
    self._windows_areas = windows_areas
    self._id = None
    self._thermal_zones = None
    self._thermal_openings = None
    self._layers = None
    self._he = ch().convective_heat_transfer_coefficient_exterior
    self._hi = ch().convective_heat_transfer_coefficient_interior
    self._u_value = None
    self._construction_name = None
    self._thickness = None
    self._internal_surface = None
    self._external_surface = None
    self._window_ratio = 0
    self._window_ratio_to_be_calculated = False
    if self._windows_areas is not None:
      self._window_ratio_to_be_calculated = True

  @property
  def id(self):
    """
    Get thermal zone id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def parent_surface(self) -> Surface:
    """
    Get the surface that belongs to the thermal boundary, considered the external surface of that boundary
    :return: Surface
    """
    return self._parent_surface

  @property
  def thermal_zones(self) -> List[ThermalZone]:
    """
    Get the thermal zones delimited by the thermal boundary
    :return: [ThermalZone]
    """
    return self._thermal_zones

  @thermal_zones.setter
  def thermal_zones(self, value):
    """
    Get the thermal zones delimited by the thermal boundary
    :param value: [ThermalZone]
    """
    self._thermal_zones = value

  @property
  def opaque_area(self):
    """
    Get the thermal boundary area in square meters
    :return: float
    """
    return float(self._opaque_area)

  @property
  def thickness(self):
    """
    Get the thermal boundary thickness in meters
    :return: float
    """
    if self._thickness is None:
      self._thickness = 0.0
      if self.layers is not None:
        for layer in self.layers:
          if not layer.no_mass:
            self._thickness += layer.thickness
    return self._thickness

  @property
  def thermal_openings(self) -> Union[None, List[ThermalOpening]]:
    """
    Get thermal boundary thermal openings
    :return: None or [ThermalOpening]
    """
    if self._thermal_openings is None:
      if self.windows_areas is not None:
        if len(self.windows_areas) > 0:
          self._thermal_openings = []
          for window_area in self.windows_areas:
            thermal_opening = ThermalOpening()
            thermal_opening.area = window_area
            self._thermal_openings.append(thermal_opening)
        else:
          self._thermal_openings = []
      else:
        if self.window_ratio is not None:
          if self.window_ratio == 0:
            self._thermal_openings = []
          else:
            thermal_opening = ThermalOpening()
            if self.window_ratio == 1:
              _area = self.opaque_area
            else:
              _area = self.opaque_area * self.window_ratio / (1-self.window_ratio)
            thermal_opening.area = _area
            self._thermal_openings = [thermal_opening]
        else:
          self._thermal_openings = []
    else:
      if self.windows_areas is not None:
        return self._thermal_openings
      if self.window_ratio is not None:
        if self.window_ratio == 0:
          self._thermal_openings = []
        else:
          if len(self._thermal_openings) == 0:
            thermal_opening = ThermalOpening()
            if self.window_ratio == 1:
              _area = self.opaque_area
            else:
              _area = self.opaque_area * self.window_ratio / (1-self.window_ratio)
            thermal_opening.area = _area
            self._thermal_openings = [thermal_opening]
          else:
            for _thermal_opening in self._thermal_openings:
              if self.window_ratio == 1:
                _area = self.opaque_area
              else:
                _area = self.opaque_area * self.window_ratio / (1-self.window_ratio)
              _thermal_opening.area = _area
              self._thermal_openings = [_thermal_opening]
    for thermal_opening in self._thermal_openings:
      thermal_opening.g_value = self._construction_archetype.window_g_value
      thermal_opening.overall_u_value = self._construction_archetype.window_overall_u_value
      thermal_opening.frame_ratio = self._construction_archetype.window_frame_ratio
      thermal_opening.construction_name = self._construction_archetype.window_type
    return self._thermal_openings

  @property
  def _construction_archetype(self):
    construction_archetypes = self.thermal_zones[0].parent_internal_zone.thermal_archetype.constructions
    for construction_archetype in construction_archetypes:
      if str(self.type) == str(construction_archetype.type):
        return construction_archetype
    return None

  @property
  def layers(self) -> List[Layer]:
    """
    Get thermal boundary layers
    :return: [Layers]
    """
    if self._construction_archetype is not None:
      self._layers = self._construction_archetype.layers
    else:
      logging.error('Layers not defined\n')
      raise ValueError('Layers not defined')
    return self._layers

  @property
  def type(self):
    """
    Get thermal boundary surface type
    :return: str
    """
    return self.parent_surface.type

  @property
  def window_ratio(self) -> Union[None, float]:
    """
    Get thermal boundary window ratio
    It returns the window ratio calculated as the total windows' areas in a wall divided by
    the total (opaque + transparent) area of that wall if windows are defined in the geometry imported.
    If not, it returns the window ratio imported from an external source (e.g. construction library, manually assigned).
    If none of those sources are available, it returns None.
    :return: float
    """
    if self._window_ratio_to_be_calculated:
      if len(self.windows_areas) == 0:
        self._window_ratio = 0
      else:
        total_window_area = 0
        for window_area in self.windows_areas:
          total_window_area += window_area
        self._window_ratio = total_window_area / (self.opaque_area + total_window_area)
    else:
      if self.type in (cte.WALL, cte.ROOF):
        if -math.sqrt(2) / 2 < math.sin(self.parent_surface.azimuth) < math.sqrt(2) / 2:
          if 0 < math.cos(self.parent_surface.azimuth):
            self._window_ratio = \
              float(self._construction_archetype.window_ratio['north']) / 100
          else:
            self._window_ratio = \
              float(self._construction_archetype.window_ratio['south']) / 100
        elif math.sqrt(2) / 2 <= math.sin(self._parent_surface.azimuth):
          self._window_ratio = \
            float(self._construction_archetype.window_ratio['east']) / 100
        else:
          self._window_ratio = \
            float(self._construction_archetype.window_ratio['west']) / 100
    return self._window_ratio

  @property
  def windows_areas(self) -> [float]:
    """
    Get windows areas
    :return: [float]
    """
    return self._windows_areas

  @property
  def u_value(self) -> Union[None, float]:
    """
    Get thermal boundary U-value in W/m2K
    internal and external convective coefficient in W/m2K values, can be configured at configuration.ini
    :return: None or float
    """
    if self._u_value is None:
      h_i = self.hi
      h_e = self.he
      if self.type == cte.GROUND:
        r_value = 1.0 / h_i + ch().soil_thickness / ch().soil_conductivity
      else:
        r_value = 1.0/h_i + 1.0/h_e
      try:
        for layer in self.layers:
          if layer.no_mass:
            r_value += float(layer.thermal_resistance)
          else:
            r_value += float(layer.thickness) / float(layer.conductivity)
        self._u_value = 1.0/r_value
      except TypeError:
        raise TypeError('Constructions layers are not initialized') from TypeError
    return self._u_value

  @property
  def construction_name(self):
    """
    Get construction name
    :return: str
    """
    if self._construction_archetype is not None:
      self._construction_name = self._construction_archetype.name
    else:
      logging.error('Construction name not defined\n')
      raise ValueError('Construction name not defined')
    return self._construction_name

  @u_value.setter
  def u_value(self, value):
    """
    Set thermal boundary U-value in W/m2K
    :param value: float
    """
    if value is not None:
      self._u_value = float(value)

  @property
  def hi(self) -> Union[None, float]:
    """
    Get internal convective heat transfer coefficient (W/m2K)
    :return: None or float
    """
    return self._hi

  @hi.setter
  def hi(self, value):
    """
    Set internal convective heat transfer coefficient (W/m2K)
    :param value: float
    """
    if value is not None:
      self._hi = value

  @property
  def he(self) -> Union[None, float]:
    """
    Get external convective heat transfer coefficient (W/m2K)
    :return: None or float
    """
    return self._he

  @he.setter
  def he(self, value):
    """
    Set external convective heat transfer coefficient (W/m2K)
    :param value: float
    """
    if value is not None:
      self._he = value

  @property
  def internal_surface(self) -> Surface:
    """
    Get the internal surface of the thermal boundary
    :return: Surface
    """
    if self._internal_surface is None:
      self._internal_surface = self.parent_surface.inverse
      # The agreement is that the layers are defined from outside to inside
      internal_layer = self.layers[len(self.layers) - 1]
      self._internal_surface.short_wave_reflectance = 1 - internal_layer.solar_absorptance
      self._internal_surface.long_wave_emittance = 1 - internal_layer.solar_absorptance

    return self._internal_surface

  @property
  def external_surface(self) -> Surface:
    if self._external_surface is None:
      # The agreement is that the layers are defined from outside to inside
      self._external_surface = self.parent_surface
      self._external_surface.short_wave_reflectance = 1 - self.layers[0].solar_absorptance
      self._external_surface.long_wave_emittance = 1 - self.layers[0].solar_absorptance
    return self._external_surface




class StringUsageToHub:
  """
  Eilat function to hub function class
  """

  def parse(self, usages) -> list[dict]:
    """
    Parse usage string in form residential-80_commercial-20
    :usages: str 
    :return: {}
    """

    parsed_usages = []
    for usage in usages.split('_'):
      usage_dict = {"usage": str(usage.split('-')[0]), "ratio": float(usage.split('-')[1])/100}
      parsed_usages.append(usage_dict)

    return parsed_usages



"""
Energy generation system (abstract class)
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Saeed Ranjbar saeed.ranjbar@concordia.ca
"""

from __future__ import annotations
from abc import ABC
from typing import Union, List

from hub.city_model_structure.energy_systems.distribution_system import DistributionSystem
from hub.city_model_structure.energy_systems.thermal_storage_system import ThermalStorageSystem
from hub.city_model_structure.energy_systems.electrical_storage_system import ElectricalStorageSystem


class GenerationSystem(ABC):
  """
  GenerationSystem class
  """
  def __init__(self):
    self._system_type = None
    self._name = None
    self._model_name = None
    self._manufacturer = None
    self._fuel_type = None
    self._distribution_systems = None
    self._energy_storage_systems = None
    self._number_of_units = None

  @property
  def system_type(self):
    """
    Get type
    :return: string
    """
    return self._system_type

  @system_type.setter
  def system_type(self, value):
    """
    Set type
    :param value: string
    """
    self._system_type = value

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return self._name

  @name.setter
  def name(self, value):
    """
    Set name
    :param value: string
    """
    self._name = value

  @property
  def model_name(self):
    """
    Get model name
    :return: string
    """
    return self._model_name

  @model_name.setter
  def model_name(self, value):
    """
    Set model name
    :param value: string
    """
    self._model_name = value

  @property
  def manufacturer(self):
    """
    Get manufacturer's name
    :return: string
    """
    return self._manufacturer

  @manufacturer.setter
  def manufacturer(self, value):
    """
    Set manufacturer's name
    :param value: string
    """
    self._manufacturer = value

  @property
  def fuel_type(self):
    """
    Get fuel_type from [Renewable, Gas, Diesel, Electricity, Wood, Coal]
    :return: string
    """
    return self._fuel_type

  @fuel_type.setter
  def fuel_type(self, value):
    """
    Set fuel_type from [Renewable, Gas, Diesel, Electricity, Wood, Coal]
    :param value: string
    """
    self._fuel_type = value

  @property
  def distribution_systems(self) -> Union[None, List[DistributionSystem]]:
    """
    Get distributions systems connected to this generation system
    :return: [DistributionSystem]
    """
    return self._distribution_systems

  @distribution_systems.setter
  def distribution_systems(self, value):
    """
    Set distributions systems connected to this generation system
    :param value: [DistributionSystem]
    """
    self._distribution_systems = value

  @property
  def energy_storage_systems(self) -> Union[None, List[ThermalStorageSystem], List[ElectricalStorageSystem]]:
    """
    Get energy storage systems connected to this generation system
    :return: [EnergyStorageSystem]
    """
    return self._energy_storage_systems

  @energy_storage_systems.setter
  def energy_storage_systems(self, value):
    """
    Set energy storage systems connected to this generation system
    :param value: [EnergyStorageSystem]
    """
    self._energy_storage_systems = value

  @property
  def number_of_units(self):
    """
    Get number of a specific generation unit
    :return: int
    """
    return self._number_of_units

  @number_of_units.setter
  def number_of_units(self, value):
    """
    Set number of a specific generation unit
    :return: int
    """
    self._number_of_units = value



"""
Usage catalog occupancy
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""
from typing import Union, List

from hub.catalog_factories.data_models.usages.schedule import Schedule


class Occupancy:
  """
  Occupancy class
  """
  def __init__(self,
               occupancy_density,
               sensible_convective_internal_gain,
               sensible_radiative_internal_gain,
               latent_internal_gain,
               schedules):
    self._occupancy_density = occupancy_density
    self._sensible_convective_internal_gain = sensible_convective_internal_gain
    self._sensible_radiative_internal_gain = sensible_radiative_internal_gain
    self._latent_internal_gain = latent_internal_gain
    self._schedules = schedules

  @property
  def occupancy_density(self) -> Union[None, float]:
    """
    Get density in persons per m2
    :return: None or float
    """
    return self._occupancy_density

  @property
  def sensible_convective_internal_gain(self) -> Union[None, float]:
    """
    Get sensible convective internal gain in Watts per m2
    :return: None or float
    """
    return self._sensible_convective_internal_gain

  @property
  def sensible_radiative_internal_gain(self) -> Union[None, float]:
    """
    Get sensible radiant internal gain in Watts per m2
    :return: None or float
    """
    return self._sensible_radiative_internal_gain

  @property
  def latent_internal_gain(self) -> Union[None, float]:
    """
    Get latent internal gain in Watts per m2
    :return: None or float
    """
    return self._latent_internal_gain

  @property
  def schedules(self) -> Union[None, List[Schedule]]:
    """
    Get occupancy schedules
    dataType = fraction
    :return: None or [Schedule]
    """
    return self._schedules

  def to_dictionary(self):
    """Class content to dictionary"""
    _schedules = []
    for _schedule in self.schedules:
      _schedules.append(_schedule.to_dictionary())
    content = {'Occupancy': {'occupancy density [persons/m2]': self.occupancy_density,
                             'sensible convective internal gain [W/m2]': self.sensible_convective_internal_gain,
                             'sensible radiative internal gain [W/m2]': self.sensible_radiative_internal_gain,
                             'latent internal gain [W/m2]': self.latent_internal_gain,
                             'schedules': _schedules}
               }
    return content



"""
Dictionaries module for Montreal system to hub energy generation system
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class MontrealGenerationSystemToHubEnergyGenerationSystem:
  """
  Montreal's generation system to hub energy generation system class
  """
  def __init__(self):
    self._dictionary = {
      'boiler': cte.BOILER,
      'furnace': cte.FURNACE,
      'cooler': cte.CHILLER,
      'electricity generator': cte.ELECTRICITY_GENERATOR,
      'photovoltaic': cte.PHOTOVOLTAIC,
      'heat pump': cte.HEAT_PUMP,
      'joule': cte.JOULE,
      'split': cte.SPLIT,
      'butane heater': cte.BUTANE_HEATER

    }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary



"""
CityObject module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

from __future__ import annotations
from typing import List, Union

from hub.city_model_structure.level_of_detail import LevelOfDetail
from hub.city_model_structure.iot.sensor import Sensor
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.attributes.polyhedron import Polyhedron

from hub.helpers.configuration_helper import ConfigurationHelper


class CityObject:
  """
  class CityObject
  """
  def __init__(self, name, surfaces):
    self._name = name
    self._level_of_detail = LevelOfDetail()
    self._surfaces = surfaces
    self._type = None
    self._city_object_lower_corner = None
    self._city_object_upper_corner = None
    self._detailed_polyhedron = None
    self._simplified_polyhedron = None
    self._min_x = ConfigurationHelper().max_coordinate
    self._min_y = ConfigurationHelper().max_coordinate
    self._min_z = ConfigurationHelper().max_coordinate
    self._max_x = ConfigurationHelper().min_coordinate
    self._max_y = ConfigurationHelper().min_coordinate
    self._max_z = ConfigurationHelper().min_coordinate
    self._centroid = None
    self._volume = None
    self._external_temperature = {}
    self._ground_temperature = {}
    self._global_horizontal = {}
    self._diffuse = {}
    self._direct_normal = {}
    self._sensors = []
    self._neighbours = None
    self._beam = {}

  @property
  def level_of_detail(self) -> LevelOfDetail:
    """
    Get level of detail of different aspects of the city: geometry, construction and usage
    :return: LevelOfDetail
    """
    return self._level_of_detail

  @property
  def name(self):
    """
    Get city object name
    :return: str
    """
    return self._name

  @property
  def type(self) -> str:
    """
    Get city object type
    :return: str
    """
    return self._type

  @property
  def volume(self) -> float:
    """
    Get city object volume in cubic meters
    :return: float
    """
    if self._volume is None:
      self._volume = self.simplified_polyhedron.volume
    return self._volume

  @volume.setter
  def volume(self, value):
    """
    Set city object volume in cubic meters
    :param value: float
    """
    self._volume = value

  @property
  def detailed_polyhedron(self) -> Polyhedron:
    """
    Get city object polyhedron including details such as holes
    :return: Polyhedron
    """
    if self._detailed_polyhedron is None:
      polygons = []
      for surface in self.surfaces:
        polygons.append(surface.solid_polygon)
        if surface.holes_polygons is not None:
          for hole_polygon in surface.holes_polygons:
            polygons.append(hole_polygon)
      self._detailed_polyhedron = Polyhedron(polygons)
    return self._detailed_polyhedron

  @property
  def simplified_polyhedron(self) -> Polyhedron:
    """
    Get city object polyhedron, just the simple lod2 representation
    :return: Polyhedron
    """
    if self._simplified_polyhedron is None:
      polygons = []
      for surface in self.surfaces:
        polygons.append(surface.perimeter_polygon)
      self._simplified_polyhedron = Polyhedron(polygons)
    return self._simplified_polyhedron

  @property
  def surfaces(self) -> List[Surface]:
    """
    Get city object surfaces
    :return: [Surface]
    """
    return self._surfaces

  @surfaces.setter
  def surfaces(self, value):
    """
    Set city object surfaces
    :return: [Surface]
    """
    self._surfaces = value

  def surface(self, name) -> Union[Surface, None]:
    """
    Get the city object surface with a given name
    :param name: str
    :return: None or Surface
    """
    for s in self.surfaces:
      if s.name == name:
        return s
    return None

  def surface_by_id(self, identification_number) -> Union[Surface, None]:
    """
    Get the city object surface with a given name
    :param identification_number: str
    :return: None or Surface
    """
    for s in self.surfaces:
      if str(s.id) == str(identification_number):
        return s
    return None

  @property
  def centroid(self) -> List[float]:
    """
    Get city object centroid
    :return: [x,y,z]
    """
    if self._centroid is None:
      self._centroid = self.simplified_polyhedron.centroid
    return self._centroid

  @property
  def max_height(self) -> float:
    """
    Get city object maximal height in meters
    :return: float
    """
    return self.simplified_polyhedron.max_z

  @property
  def external_temperature(self) -> {float}:
    """
    Get external temperature surrounding the city object in Celsius
    :return: dict{dict{[float]}}
    """
    return self._external_temperature

  @external_temperature.setter
  def external_temperature(self, value):
    """
    Set external temperature surrounding the city object in Celsius
    :param value: dict{dict{[float]}}
    """
    self._external_temperature = value

  @property
  def ground_temperature(self) -> dict:
    """
    Get ground temperature under the city object in Celsius at different depths in meters for different time steps
    example of use: {month: {0.5: [10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]}}
    :return: dict{dict{[float]}}
    """
    return self._ground_temperature

  @ground_temperature.setter
  def ground_temperature(self, value):
    """
    Set ground temperature under the city object in Celsius at different depths
    :param value: dict{dict{[float]}}
    """
    self._ground_temperature = value

  @property
  def global_horizontal(self) -> dict:
    """
    Get global horizontal radiation surrounding the city object in J/m2
    :return: dict{dict{[float]}}
    """
    return self._global_horizontal

  @global_horizontal.setter
  def global_horizontal(self, value):
    """
    Set global horizontal radiation surrounding the city object in J/m2
    :param value: dict{dict{[float]}}
    """
    self._global_horizontal = value

  @property
  def diffuse(self) -> dict:
    """
    Get diffuse radiation surrounding the city object in J/m2
    :return: dict{dict{[float]}}
    """
    return self._diffuse

  @diffuse.setter
  def diffuse(self, value):
    """
    Set diffuse radiation surrounding the city object in J/m2
    :param value: dict{dict{[float]}}
    """
    self._diffuse = value

  @property
  def direct_normal(self) -> dict:
    """
    Get beam radiation surrounding the city object in J/m2
    :return: dict{dict{[float]}}
    """
    return self._direct_normal

  @direct_normal.setter
  def direct_normal(self, value):
    """
    Set beam radiation surrounding the city object in J/m2
    :param value: dict{dict{[float]}}
    """
    self._direct_normal = value

  @property
  def lower_corner(self):
    """
    Get city object lower corner coordinates [x, y, z]
    :return: [x,y,z]
    """
    if self._city_object_lower_corner is None:
      self._city_object_lower_corner = [self._min_x, self._min_y, self._min_z]
    return self._city_object_lower_corner

  @property
  def upper_corner(self):
    """
    Get city object upper corner coordinates [x, y, z]
    :return: [x,y,z]
    """
    if self._city_object_upper_corner is None:
      self._city_object_upper_corner = [self._max_x, self._max_y, self._max_z]
    return self._city_object_upper_corner

  @property
  def sensors(self) -> List[Sensor]:
    """
    Get sensors belonging to the city object
    :return: [Sensor]
    """
    return self._sensors

  @sensors.setter
  def sensors(self, value):
    """
    Set sensors belonging to the city object
    :param value: [Sensor]
    """
    self._sensors = value

  @property
  def neighbours(self) -> Union[None, List[CityObject]]:
    """
    Get the list of neighbour_objects and their properties associated to the current city_object
    """
    return self._neighbours

  @neighbours.setter
  def neighbours(self, value):
    """
    Set the list of neighbour_objects and their properties associated to the current city_object
    """
    self._neighbours = value

  @property
  def beam(self) -> dict:
    """
    Get beam radiation surrounding the city object in J/m2
    :return: dict{dict{[float]}}
    """
    return self._beam

  @beam.setter
  def beam(self, value):
    """
    Set beam radiation surrounding the city object in J/m2
    :param value: dict{dict{[float]}}
    """
    self._beam = value



"""
Household module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class Household:
  """
  Household class
  """

  def __init__(self):
    self._number_of_people = None
    self._number_of_cars = None

  @property
  def number_of_people(self):
    """
    Get number of people leaving in the household
    :return: int
    """
    return self._number_of_people

  @property
  def number_of_cars(self):
    """
    Get number of cars owned by the householders
    :return: int
    """
    return self._number_of_cars



"""
City Object repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez Guillermo.GutierrezMorote@concordia.ca
"""
import datetime
import logging
from typing import Union

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from hub.city_model_structure.building import Building
from hub.persistence.models import CityObject as Model
from hub.persistence.models import City as CityModel
from hub.persistence.repository import Repository


class CityObject(Repository):
  """
  City object repository
  """
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(CityObject, cls).__new__(cls)
    return cls._instance

  def insert(self, city_id: int, building: Building):
    """
    Inserts a new city object
    :param city_id: city id for the city owning this city object
    :param building: the city object (only building for now) to be inserted
    return Identity id
    """
    city_object = self.get_by_name_or_alias_and_city(building.name, city_id)
    if city_object is not None:
      raise SQLAlchemyError(f'A city_object named {building.name} already exists in that city')
    try:
      city_object = Model(city_id=city_id,
                          building=building)
      with Session(self.engine) as session:
        session.add(city_object)
        session.flush()
        session.commit()
        session.refresh(city_object)
      return city_object.id
    except SQLAlchemyError as err:
      logging.error('An error occurred while creating city_object %s', err)
      raise SQLAlchemyError from err

  def update(self, city_id: int, building: Building):
    """
    Updates an application
    :param city_id: the city id of the city owning the city object
    :param building: the city object
    :return: None
    """
    try:
      object_usage = ''
      for internal_zone in building.internal_zones:
        for usage in internal_zone.usages:
          object_usage = f'{object_usage}{usage.name}_{usage.percentage} '
      object_usage = object_usage.rstrip()
      with Session(self.engine) as session:
        session.query(Model).filter(Model.name == building.name, Model.city_id == city_id).update(
          {'name': building.name,
           'aliases': building.aliases,
           'object_type': building.type,
           'year_of_construction': building.year_of_construction,
           'function': building.function,
           'usage': object_usage,
           'volume': building.volume,
           'area': building.floor_area,
           'updated': datetime.datetime.utcnow()})
        session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while updating city object %s', err)
      raise SQLAlchemyError from err

  def delete(self, city_id: int, name: str):
    """
    Deletes an application with the application_uuid
    :param city_id: The id for the city owning the city object
    :param name: The city object name
    :return: None
    """
    try:
      with Session(self.engine) as session:
        session.query(Model).filter(Model.city_id == city_id, Model.name == name).delete()
        session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while deleting application %s', err)
      raise SQLAlchemyError from err

  def building_in_cities_info(self, name, cities):
    """
    Fetch a city object based on name and city id
    :param name: city object name
    :param cities: city identifiers
    :return: [CityObject] with the provided name or alias belonging to the city with id city_id
    """
    try:
      # search by name first
      with Session(self.engine) as session:
        city_object = session.execute(select(Model).where(
          Model.name == name, Model.city_id.in_(cities))
        ).first()
        if city_object is not None:
          return city_object[0]
        # name not found, so search by alias instead
        city_objects = session.execute(
          select(Model).where(Model.aliases.contains(name), Model.city_id.in_(cities))
        ).all()
      for city_object in city_objects:
        aliases = city_object[0].aliases.replace('{', '').replace('}', '').split(',')
        for alias in aliases:
          if alias == name:
            # force the name as the alias
            city_object[0].name = name
            return city_object[0]
      return None
    except SQLAlchemyError as err:
      logging.error('Error while fetching city object by name and city: %s', err)
      raise SQLAlchemyError from err
    except IndexError as err:
      logging.error('Error while fetching city object by name and city, empty result %s', err)
      raise IndexError from err

  def get_by_name_or_alias_for_user_app(self, user_id, application_id, names) -> Union[Model, None]:
    """
    Fetch city objects belonging to the user and application where the name or alias is in the names list
    :param user_id: User ID
    :param application_id: Application ID
    :param names: a list of building aliases or names
    :return [CityObject] or None
    """
    with Session(self.engine) as session:
      cities = session.execute(select(CityModel).where(
        CityModel.user_id == user_id, CityModel.application_id == application_id
      )).all()
      ids = [c[0].id for c in cities]
      buildings = session.execute(select(Model).where(
        Model.city_id.in_(ids), Model.name.in_(names)
      ))
      results = [r[0] for r in buildings]
      print(ids, buildings)
    return None

  def get_by_name_or_alias_and_city(self, name, city_id) -> Union[Model, None]:
    """
    Fetch a city object based on name and city id
    :param name: city object name
    :param city_id: a city identifier
    :return: [CityObject] with the provided name or alias belonging to the city with id city_id
    """
    try:
      # search by name first
      with Session(self.engine) as session:
        city_object = session.execute(select(Model).where(Model.name == name, Model.city_id == city_id)).first()
        if city_object is not None:
          return city_object[0]
        # name not found, so search by alias instead
        city_objects = session.execute(
          select(Model).where(Model.aliases.contains(name), Model.city_id == city_id)
        ).all()
      for city_object in city_objects:
        aliases = city_object[0].aliases.replace('{', '').replace('}', '').split(',')
        for alias in aliases:
          if alias == name:
            # force the name as the alias
            city_object[0].name = name
            return city_object[0]
      return None
    except SQLAlchemyError as err:
      logging.error('Error while fetching city object by name and city: %s', err)
      raise SQLAlchemyError from err
    except IndexError as err:
      logging.error('Error while fetching city object by name and city, empty result %s', err)
      raise IndexError from err

  def get_by_name_or_alias_in_cities(self, name, city_ids) -> Model:
    """
    Fetch a city object based on name and city ids
    :param name: city object name
    :param city_ids: a list of city identifiers
    :return: [CityObject] with the provided name or alias belonging to the city with id city_id
    """
    try:
      # search by name first
      with Session(self.engine) as session:
        city_object = session.execute(select(Model).where(Model.name == name, Model.city_id.in_(tuple(city_ids)))).first()
        if city_object is not None:
          return city_object[0]
        # name not found, so search by alias instead
        city_objects = session.execute(
          select(Model).where(Model.aliases.contains(name), Model.city_id.in_(tuple(city_ids)))
        ).all()
      for city_object in city_objects:
        aliases = city_object[0].aliases.replace('{', '').replace('}', '').split(',')
        for alias in aliases:
          if alias == name:
            # force the name as the alias
            city_object[0].name = name
            return city_object[0]
      return None
    except SQLAlchemyError as err:
      logging.error('Error while fetching city object by name and city: %s', err)
      raise SQLAlchemyError from err
    except IndexError as err:
      logging.error('Error while fetching city object by name and city, empty result %s', err)
      raise IndexError from err



"""
Dictionaries module for Montreal system catalog demand types to hub energy demand types
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import hub.helpers.constants as cte


class NorthAmericaDemandTypeToHubEnergyDemandType:
  """
  Montreal demand type to hub energy demand type
  """
  def __init__(self):
    self._dictionary = {'heating': cte.HEATING,
                        'cooling': cte.COOLING,
                        'domestic_hot_water': cte.DOMESTIC_HOT_WATER,
                        'electricity': cte.ELECTRICITY,
                        }

  @property
  def dictionary(self) -> dict:
    """
    Get the dictionary
    :return: {}
    """
    return self._dictionary
